#  Copyright 2023 Liga Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Utilities to log sklearn models to MLflow for Liga
"""

from typing import Any, Optional

from sklearn.base import (
    RegressorMixin,
    ClassifierMixin,
    TransformerMixin,
    ClusterMixin,
)


def _get_model_type(model: Any) -> str:
    if isinstance(model, RegressorMixin):
        return "liga.sklearn.models.regressor"
    elif isinstance(model, ClassifierMixin):
        return "liga.sklearn.models.classifier"
    elif isinstance(model, ClusterMixin):
        if "predict" in dir(model):
            return "liga.sklearn.models.cluster"
        else:
            raise RuntimeError(
                "Clustering without predict method is not supported"
            )
    elif isinstance(model, TransformerMixin):
        return "liga.sklearn.models.transformer"
    else:
        raise RuntimeError("No corresponding ModelType yet")


def log_model(
    model: Any,
    registered_model_name: Optional[str] = None,
    schema: Optional[str] = None,
    labels: Optional[dict] = None,
    **kwargs: Any,
) -> None:
    """
    Log sklearn models to MLflow with proper default settings
    """
    model_type = _get_model_type(model)

    from liga.mlflow import logger

    logger.sklearn.log_model(  # type: ignore[attr-defined]
        model,
        "model",
        schema,
        registered_model_name,
        "liga.sklearn",
        model_type,
        labels,
        **kwargs,
    )



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
ModelType: Regressor
"""

from typing import List, Any

from liga.sklearn.models import SklearnModelType


class Regressor(SklearnModelType):
    """
    Regression models.
    """

    def schema(self) -> str:
        return "float"

    def predict(self, *args: Any, **kwargs: Any) -> List[float]:
        assert self.model is not None
        assert len(args) == 1
        return self.model.predict(args[0]).tolist()


MODEL_TYPE = Regressor()



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Generate UDF for Liga and sklearn integration
"""

import pickle
from typing import Iterator, Any

import numpy as np
import pandas as pd
from pyspark.serializers import CloudPickleSerializer
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import BinaryType

from liga.io import open_uri
from liga.registry.base import ModelSpec

__all__ = ["generate_udf"]

_pickler = CloudPickleSerializer()


def generate_udf(spec: ModelSpec):  # type: ignore[no-untyped-def]
    """Construct a UDF to run sklearn model.

    Parameters
    ----------
    spec : ModelSpec
        the model specifications object

    Returns
    -------
    A Spark Pandas UDF.
    """

    def sklearn_inference_udf(
        iterator: Iterator[pd.Series],
    ) -> Iterator[pd.Series]:
        model = spec.model_type
        model.load_model(spec)
        for series in iterator:
            X = np.vstack(series.apply(_pickler.loads).to_numpy())
            y = [_pickler.dumps(pred) for pred in model.predict(X)]
            yield pd.Series(y)

    return pandas_udf(sklearn_inference_udf, returnType=BinaryType())  # type: ignore[call-overload]


def load_model_from_uri(uri: str) -> Any:
    """Load a Sklearn model from URL.

    Assuming model is persisted via pickle following the instruction:
    https://scikit-learn.org/stable/model_persistence.html

    """
    with open_uri(uri) as fobj:
        return pickle.load(fobj)




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: //:liga-sklearn

from setuptools import setup

setup(**{
    'install_requires': (
        'liga==0.2.0.dev11',
        'scikit-learn',
    ),
    'name': 'liga-sklearn',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'liga.sklearn',
        'liga.sklearn.models',
    ),
    'python_requires': '<3.10,>=3.8',
    'version': '0.2.0.dev11',
})
